import numpy as np
from skimage import measure
import warnings
import csv


# limiting the precision to 10 digits seems reasonable
# we do this in order to get rid of double-precision artifacts
def _formatter(x):
    if isinstance(x, float):
        return "%.10g" % x
    return x


def write_dict_to_csv(dictionary: dict, filename: str, delimiter: str = ",") -> None:
    """
     Write a dictionary to a text file.

     Parameters
     ----------
         dictionary : dict
            The dictionary to be written.
         filename : str
            The name of the output file.
         delimiter : str
            The delimiter to use between fields (default is ",").

     Returns:
         None
    """

    with open(filename, "w") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=[key for key in dictionary], delimiter=delimiter)
        writer.writeheader()
        # use the formatter to convert floating point numbers into a string
        for i in range(len(dictionary[list(dictionary.keys())[0]])):
            writer.writerow({key: _formatter(dictionary[key][i]) for key in dictionary})


def _select_isoline(isoline, search_pos):
    idx = [index for index, array in enumerate(isoline) if np.any((array.T)[0] == search_pos)]
    if len(idx) == 0:
        if len(isoline) < 2:
            isoline = (np.squeeze(isoline)).T
        else:
            isoline = (max(isoline, key=lambda x: len(x))).T
    elif len(idx) == 1:
        # One isoline fullfills the criterion -> we're done
        isoline = isoline[idx[0]].T
    else:
        # Multiple isolines fullfill the criterion -> use largest one
        isoline_red = []
        for index in idx:
            isoline_red.append(isoline[idx[index]])
        isoline = (max(isoline_red, key=lambda x: len(x))).T
    return isoline


def extract_isolines(simdata: dict, frame: int):
    """
     Extracts three distinct isolines for a given frame index

        Parameters
        ----------
        simdata : dict
            Simulation data containing a list of numpy arrays for keys "phia", "phib" and "phic"
        frame : int
            Frame index (zero based)

        Returns
        -------
        iso_ab, iso_ac, iso_bc
            Three isolines phia == phib, phia == phic, phib == phic
    """

    phia = simdata["phia"][frame]
    phib = simdata["phib"][frame]
    phic = simdata["phic"][frame]

    iso_ab = measure.find_contours((phia-phib), 0.0)
    iso_ac = measure.find_contours((phia-phic), 0.0)
    iso_bc = measure.find_contours((phib-phic), 0.0)

    # The marching squares might return a list with multiple disconnected lines
    # and we try to extract the desired one.
    # Isoline ac should have a x=0 entry
    iso_ac = _select_isoline(iso_ac, 0)

    # Isoline bc should have a x=Nx-1 entry
    iso_bc = _select_isoline(iso_bc, phib.shape[0] - 1)

    # For iso_ab we don't have a good criterion and this isoline is less
    # important -> use largest list entry
    if len(iso_ab) < 2:
        iso_ab = (np.squeeze(iso_ab)).T
    else:
        iso_ab = (max(iso_ab, key=lambda x: len(x))).T

    # NOTE: assumes STRUCTURED_POINTS with equidistant spacings
    # offset and delta x in dimensionless units
    x_offset = float(simdata["grid"]["origin_x"] / simdata["input_params"]["width"])
    y_offset = float(simdata["grid"]["origin_y"] / simdata["input_params"]["width"])
    dx = float(simdata["grid"]["delta_x"] / simdata["input_params"]["width"])
    dy = float(simdata["grid"]["delta_y"] / simdata["input_params"]["width"])

    if iso_ab.size != 0:
        iso_ab = np.stack((iso_ab[0] * dx + x_offset, iso_ab[1]*dy+y_offset), axis=1).T
    if iso_ac.size != 0:
        iso_ac = np.stack((iso_ac[0] * dx + x_offset, iso_ac[1] * dy + y_offset), axis=1).T
    if iso_bc.size != 0:
        iso_bc = np.stack((iso_bc[0] * dx + x_offset, iso_bc[1] * dy + y_offset), axis=1).T

    return iso_ab, iso_ac, iso_bc


def _split_half(x, y):
    """ Splits two given one-dimensional arrays of x- and y- coordinates by half"""

    if len(x) != 2:
        x_split = np.array_split(x, 2)
        y_split = np.array_split(y, 2)
        # duplicate node at which the array is split
        x_split[1] = np.concatenate(([x_split[0][-1]], x_split[1]))
        y_split[1] = np.concatenate(([y_split[0][-1]], y_split[1]))
    else:
        x_split = np.array([x])
        y_split = np.array([y])
    return x_split, y_split


def _is_overlap_1d(x1, x2) -> bool:
    """Checks if two one-dimensional arrays have overlapping intervals"""

    min_x1, min_x2 = np.min(x1), np.min(x2)
    max_x1, max_x2 = np.max(x1), np.max(x2)
    return not (max_x1 < min_x2 or max_x2 < min_x1)


def _is_overlap(x1, y1, x2, y2) -> bool:
    """Checks if two contours have overlapping intervals"""

    return (_is_overlap_1d(x1, x2) and _is_overlap_1d(y1, y2))


def _solve_intersect(segment1, segment2):
    """
    Calculates the intersection between two straight line-segments (each given by a start and end point)

        Parameters
        ----------
        segment1 : Two dimensional array
            Containg start (0-index) and end (1-index) points of first segment
        segment2 : Two dimensional array
            Containg start (0-index) and end (1-index) points of second segment

        Returns
        -------
        numpy array
            Point of intersection
    """

    x0 = np.array(segment1[0])
    t0 = np.array(segment1[1]) - np.array(segment1[0])
    # normalize
    length_0 = np.linalg.norm(t0)
    t0 = t0 / length_0

    x1 = np.array(segment2[0])
    t1 = np.array(segment2[1]) - np.array(segment2[0])
    # normalize
    length_1 = np.linalg.norm(t1)
    t1 = t1 / length_1
    # rhs
    r0 = np.dot(t0, x1-x0)
    r1 = np.dot(t1, x1-x0)

    t0t1 = np.dot(t0, t1)

    det_M = 1.0 - t0t1*t0t1
    # both segments are colinear? - intersection ill defined
    if abs(det_M) < 1.0e-12:
        return None

    lambda_0 = (r0 - r1*t0t1) / det_M
    lambda_1 = (r0*t0t1 - r1) / det_M

    # check if intersection is within the bounds of the sections -> none if outside
    if (lambda_0 < 0.0 or lambda_0 > length_0):
        return None
    if (lambda_1 < 0.0 or lambda_1 > length_1):
        return None

    # interpolate between the two nodes
    return x0 + lambda_0*t0


def _calc_intersection(x1, y1, x2, y2, intersections) -> None:
    """Recursively calculates intersections for two given contours"""

    if len(x1) == 2 and len(x2) == 2:
        segment1 = [[x1[0], y1[0]], [x1[1], y1[1]]]
        segment2 = [[x2[0], y2[0]], [x2[1], y2[1]]]
        intersect = _solve_intersect(segment1, segment2)
        if intersect is not None:
            if len(intersections) != 0:
                if not np.allclose(intersect, intersections[-1]):
                    intersections.append(intersect)
            else:
                intersections.append(intersect)
    else:
        # one of the curves is not yet a linear segment
        # hence we split them by half
        x1_parts, y1_parts = _split_half(x1, y1)
        x2_parts, y2_parts = _split_half(x2, y2)
        # figure out which of the parts overlap
        for i in range(len(x1_parts)):
            for j in range(len(x2_parts)):
                # if they overlap, there may be intersections
                if _is_overlap(x1_parts[i], y1_parts[i], x2_parts[j], y2_parts[j]):
                    _calc_intersection(x1_parts[i], y1_parts[i], x2_parts[j], y2_parts[j], intersections)


def calc_intersection(x1, y1, x2, y2):
    """
    Calculates intersections between two given contours.

        Parameters
        ----------
        x1 : array_like
            x-coordinates for first contour
        y1 : array_like
            y-coordinates for first contour
        x2 : array_like
            x-coordinates for second contour
        y2 : array_like
            y-coordinates for second contour

        Returns
        -------
        numpy array
            Two-dimensional array containing x- (0-index) and y-coordinates (1-index)
    """
    if len(y1) != len(x1):
        raise ValueError("Size of x1 unequal to y1")
    if len(x2) != len(y2):
        raise ValueError("Size of x2 unequal to y2")

    intersections = []
    _calc_intersection(x1, y1, x2, y2, intersections)
    return np.array([p[0] for p in intersections]), np.array([p[1] for p in intersections])


def _phase_fraction(phase_field):
    return np.sum(phase_field)/(phase_field.size)


def calculate_velocity_nondimensional(simdata: dict, frame1: int, frame2: int):
    """
    Compute velocity of the grain boundary from isolines between phase-fields.
    This function returns non-dimensional velocities.
    Computation is based on backward Euler difference scheme.

    Parameters
    ----------
    simdata : dict
        Simulation data
    frame1 : int
        Start frame index
    frame2 : int
        End frame index

    Returns
    -------
    dict
        4 velocities which are differently computed
        "left_boundary": phia=phic isoline at left boundary,
        "right_boundary": phib=phic isoline at right boundary,
        "triple_point": calculated from intersection of isolines phia==phic and phib==phic,
        "phase_fraction": based on phase evolution of phic.
    """

    if frame1 > frame2:
        print("Frame 1 must be smaller than frame 2 for correct computation of velocity.")
        return None

    times = []
    ymax_left = []
    ymax_right = []
    tp_xy = []
    fraction_c = []

    for frame in [frame1, frame2]:
        iso_ab, iso_ac, iso_bc = extract_isolines(simdata, frame)

        # Check if some isolines are empty sets
        if iso_ac.size != 0:
            ymax_left.append(np.max(iso_ac[1]))
        else:
            ymax_left.append(np.nan)
            warnings.warn("Isoline iso_ac is an empty set!")
        if iso_bc.size != 0:
            ymax_right.append(np.max(iso_bc[1]))
        else:
            ymax_right.append(np.nan)
            warnings.warn("Isoline iso_bc is an empty set!")

        if iso_ac.size != 0 and iso_bc.size != 0:
            # Compute triple point position
            tp_x, tp_y = calc_intersection(iso_ac[0], iso_ac[1], iso_bc[0], iso_bc[1])
            # Check if isolines intersect or yield empty set
            if tp_x.size != 0:
                tp_xy.append([tp_x[0], tp_y[0]])
            else:
                warnings.warn("No triple point found!")
                tp_xy.append([np.nan, np.nan])
        else:
            tp_xy.append([np.nan, np.nan])
        # Compute phase fraction of phic
        fraction_c.append(_phase_fraction(simdata["phic"][frame]))
        times.append(simdata["time"][frame])

    input_params = simdata["input_params"]
    factor = float((input_params["width"]**2 / (input_params["gb_mobility"] * input_params["gb_energy_horizontal"])) / (times[1] - times[0]))

    velocity = {}
    velocity["left_boundary"] = (ymax_left[1] - ymax_left[0]) * factor
    velocity["right_boundary"] = (ymax_right[1] - ymax_right[0]) * factor
    velocity["triple_point"] = (tp_xy[1][1] - tp_xy[0][1]) * factor

    dy = float(simdata["grid"]["delta_y"] / simdata["input_params"]["width"])
    velocity["phase_fraction"] = -(fraction_c[1]-fraction_c[0]) * (simdata["phic"][0].shape[1]*dy) * factor

    return velocity


def extract_grainboundary_profile(simdata: dict, frame: int):
    """
    Compute the grain boundary profile from isolines between phase-fields.

    Parameters
    ----------
    simdata : dict
        Simulation data
    frame : int
        Frame index (0-based)

    Returns
    -------
    x, y
        Two one-dimensional arrays containing the dimensionless x- and y-coordinates of the profile.
    """

    _, iso_ac, iso_bc = extract_isolines(simdata, frame)

    if iso_ac.size != 0 and iso_bc.size != 0:
        # dimensionless delta x
        dx = float(simdata["grid"]["delta_x"] / simdata["input_params"]["width"])

        # Compute triple point position
        tp_x, tp_y = calc_intersection(iso_ac[0], iso_ac[1], iso_bc[0], iso_bc[1])
        # Check if isolines intersect or yield empty set
        if tp_x.size == 0:
            warnings.warn("Triple point is ill-defined!")
            # Assume tp_x is in the middle of simulation domain
            tp_x = 0.5

        xs = np.concatenate((iso_ac[0, iso_ac[0, :] < (tp_x - 0.01*dx)],
                             iso_bc[0, iso_bc[0, :] > (tp_x + 0.01*dx)]), axis=0)
        ys = np.concatenate((iso_ac[1, iso_ac[0, :] < (tp_x - 0.01*dx)],
                             iso_bc[1, iso_bc[0, :] > (tp_x + 0.01*dx)]), axis=0)

        if xs.size != 0:
            # Only return values that are inside the physical domain
            ys = ys[(xs >= 0.0) & (xs <= 1.0)]
            xs = xs[(xs >= 0.0) & (xs <= 1.0)]

            # Shift such that highest point is at y=0
            ys -= np.max(ys)
        else:
            warnings.warn("Something fishy here!")
            xs = np.array([-1])
            ys = [np.nan]

    else:
        warnings.warn("Isoline iso_ac or iso_bc is an empty set!")
        xs = np.array([-1])
        ys = np.nan

    return xs, ys



import numpy as np
import triplepy.sharp_interface_solution as si


def test_si_velocity():
    driving_forces = np.array([0.1, 1.0, 10.0, 100.0])
    driving_forces = np.concatenate((-np.flip(driving_forces), [0.0], driving_forces))
    # Solutions taken from Wolfram Mathematica
    velocity_solutions_neg = [-115.4700538379255, -11.79727890422015, -2.098316790238965, -1.152126956202861]
    velocity_solution_0 = [-1.047197551196597]
    velocity_solutions_pos = [-0.9423073427275777, 0.0, 9.297707785163341, 99.83162536999436]
    solutions = np.concatenate((velocity_solutions_neg, velocity_solution_0, velocity_solutions_pos))
    calculator = si.GB_VelocityCalculator(si.calculate_slope(1.0, 1.0))
    assert np.allclose(calculator.calculate_velocity(driving_forces), solutions)



import numpy as np
import triplepy.sharp_interface_solution as analytic
from triplepy.numerical_evaluation import write_dict_to_csv

angles_deg = np.array([60.0, 90.0, 120.0, 150.0])
slopes = analytic.calculate_slope_from_dihedral_angle(np.deg2rad(angles_deg))
ratios = analytic.calculate_gb_energy_ratio_from_dihedral_angle(np.deg2rad(angles_deg))

for i, m in enumerate(slopes):
    # Compute slope from Young's law
    print("Dihedral angle = %.6g" % angles_deg[i])
    print("TJ slope = %.6g" % m)
    print("Energy ratio = %.6g" % ratios[i])

    # Compute non-dimensional driving force
    p = np.linspace(-10, 10, num=501)

    # Compute (non-dimensional) analytical velocity
    gb = analytic.GB_VelocityCalculator(m)
    v_analytic = gb.calculate_velocity(p)

    result_dict = {"dimless_drivingforce": p, "v_analytic": v_analytic, "sqrt1m2_p": np.sqrt(1.0 + m*m) * p}
    write_dict_to_csv(result_dict, "v_analytic_theta%.10g.csv" % angles_deg[i])
    print("------------------------")



import numpy as np
import triplepy.sharp_interface_solution as si


def test_stationary_gb():
    # we check 4 different aspect ratios
    gamma_ratios = np.array([0.5, 1.0, 1.5, 1.8])
    # corresponding triple-junction slopes
    tj_slopes = si.calculate_slope(1.0, gamma_ratios)
    # corresponding dihedral angles
    tj_angles = si.calculate_dihedral_angle(1.0, gamma_ratios)
    # the balancing driving force
    driving_forces = 2.0 * tj_slopes / np.sqrt(1.0 + tj_slopes**2)
    # the height-to-width ratio for the circular arc passing through the given triple junctions
    hw_ratios_analytic = (1.0 - np.sin(0.5 * tj_angles)) / (2.0*np.cos(0.5 * tj_angles))
    hw_ratios = []
    rtol = 1.0e-8
    for i, m in enumerate(tj_slopes):
        solver = si.GB_GeometrySolver(m, driving_forces[i])
        geometry = solver.calc_dimensionless_geometry(kind="double", relative_l2_tolerance=rtol)
        width = np.max(geometry["x"]) - np.min(geometry["x"])
        height = np.max(geometry["y"]) - np.min(geometry["y"])
        hw_ratios.append(height / width)
    # we check 1 digit less due to rounding
    assert np.allclose(hw_ratios_analytic, hw_ratios, rtol=10.0 * rtol)


def test_capillary_only_gb():
    # we check 4 different aspect ratios
    gamma_ratios = np.array([0.5, 1.0, 1.5, 1.8])
    # corresponding triple-junction slopes
    tj_slopes = si.calculate_slope(1.0, gamma_ratios)
    # corresponding dihedral angles
    tj_angles = si.calculate_dihedral_angle(1.0, gamma_ratios)
    # the height-to-width ratio according to Eq.~(42) in https://doi.org/10.1016/j.commatsci.2022.111995
    hw_ratios_analytic = np.log(np.sin(0.5 * tj_angles)) / (tj_angles - np.pi)
    hw_ratios = []
    rtol = 1.0e-8
    for m in tj_slopes:
        solver = si.GB_GeometrySolver(m, 0.0)
        geometry = solver.calc_dimensionless_geometry(kind="double", relative_l2_tolerance=rtol)
        width = np.max(geometry["x"]) - np.min(geometry["x"])
        height = np.max(geometry["y"]) - np.min(geometry["y"])
        hw_ratios.append(height / width)
    # we check 1 digit less due to rounding
    assert np.allclose(hw_ratios_analytic, hw_ratios, rtol=10.0 * rtol)


def calc_differential_eq(y_prime, x_spacing, p, v):
    # we test the ODE in the following form which allows to use a relative error on unity
    # y''(x) / (v (1 + (y'(x))^2) - p (1 + (y'(x))^2)^(3/2)) = 1
    y_2prime = np.gradient(y_prime, x_spacing, edge_order=2)
    sqrt_term = np.sqrt(1.0 + y_prime**2)
    return y_2prime / (v*(sqrt_term**2) - p*sqrt_term**3)


def test_general_gbs():
    # we check 4 different aspect ratios
    gamma_ratios = np.array([0.5, 1.0, 1.5, 1.8])
    # corresponding triple-junction slopes
    tj_slopes = si.calculate_slope(1.0, gamma_ratios)

    for m in tj_slopes:
        for p in [-2.0, -1.0, -0.5, 0.5, 1.0, 2.0]:
            calc = si.GB_VelocityCalculator(m)
            v = calc.calculate_velocity(p)
            solver = si.GB_GeometrySolver(m, p)
            geometry = solver.calc_dimensionless_geometry(kind="half", relative_l2_tolerance=1e-8)

            # check that the derivative is correct (three digits)
            assert np.allclose(np.gradient(geometry["y"], geometry["x"], edge_order=2), geometry["derivative"], rtol=1e-3)

            # check that ODE is satisfied
            x_min = np.min(geometry["x"])
            x_max = np.max(geometry["x"])
            n_points = len(geometry["x"]) // 2

            x_spacing = (x_max - x_min) / n_points
            x_equidistant = np.linspace(x_min, x_max, n_points)
            yprime = np.interp(x_equidistant, geometry["x"], geometry["derivative"])
            values = calc_differential_eq(yprime, x_spacing, p, v)

            # the accuracy with interpolation is limited to three digits
            assert np.allclose(values, 1.0, rtol=1e-3)



import os
import vtk
from vtk.util.numpy_support import vtk_to_numpy
import json
from xml.dom import minidom
import numpy as np
import warnings
from triplepy.set_input_parameters import load_sim_params_from_json


def read_vtk_header(filename):
    header = {}

    if filename.endswith(".vtk"):
        reader = vtk.vtkStructuredPointsReader()
        reader.SetFileName(filename)
        reader.ReadAllScalarsOn()
    else:
        # vti format
        reader = vtk.vtkXMLImageDataReader()
        reader.SetFileName(filename)

    reader.Update()
    data = reader.GetOutput()

    header["origin"] = data.GetOrigin()
    header["spacing"] = data.GetSpacing()

    return header


def read_vtk_field_data(filename):
    fields = {"point_data": {}, "cell_data": {}}

    if filename.endswith(".vtk"):
        reader = vtk.vtkStructuredPointsReader()
        reader.SetFileName(filename)
        reader.ReadAllScalarsOn()
    else:
        # vti format
        reader = vtk.vtkXMLImageDataReader()
        reader.SetFileName(filename)

    reader.Update()
    data = reader.GetOutput()

    # read all point data
    for i in range(data.GetPointData().GetNumberOfArrays()):
        array = data.GetPointData().GetAbstractArray(i)
        fields["point_data"][array.GetName()] = vtk_to_numpy(array).reshape(data.GetDimensions(), order="F").squeeze()

    # read all cell data
    for i in range(data.GetCellData().GetNumberOfArrays()):
        cell_dimensions = [n - 1 if n != 1 else n for n in data.GetDimensions()]
        array = data.GetCellData().GetAbstractArray(i)
        fields["cell_data"][array.GetName()] = vtk_to_numpy(array).reshape(cell_dimensions, order="F").squeeze()

    return fields


def read_vtk_file(filename):
    print("Reading file \"%s\"" % filename)
    result_dict = {}
    header_dict = read_vtk_header(filename)
    fields_dict = read_vtk_field_data(filename)
    result_dict = {**header_dict, **fields_dict}

    return result_dict


def read_vtk_time_series(series_filename):
    simpath = os.path.dirname(series_filename)

    with open(series_filename) as json_file:
        vtk_series = json.load(json_file)

    frames = {"time": [], "data": []}
    for vtk_file in vtk_series["files"]:
        filename = vtk_file["name"]
        if not os.path.isabs(filename):
            filename = os.path.join(simpath, filename)

        vtk_data = read_vtk_file(filename)
        frames["data"].append(vtk_data)
        frames["time"].append(float(vtk_file["time"]))
    return frames


def read_pvd_time_series(pvd_filename):
    simpath = os.path.dirname(pvd_filename)

    doc = minidom.parse(pvd_filename)
    vtkfile = doc.getElementsByTagName("VTKFile")[0]
    collection = vtkfile.getElementsByTagName("Collection")[0]
    datasets = collection.getElementsByTagName("DataSet")

    frames = {"time": [], "data": []}
    for dataset in datasets:
        if not dataset.hasAttribute("timestep"):
            raise IOError("File \"%s\" has a missing timestep" % pvd_filename)

        if not dataset.hasAttribute("file"):
            raise IOError("File \"%s\" has a missing file" % pvd_filename)

        if not dataset.hasAttribute("part"):
            raise IOError("File \"%s\" has a missing part" % pvd_filename)

        if int(dataset.getAttribute("part")) != 0:
            raise IOError("Only a single part supported in pvd file")

        filename = dataset.getAttribute("file")
        if not os.path.isabs(filename):
            filename = os.path.join(simpath, filename)

        vtk_data = read_vtk_file(filename)
        frames["data"].append(vtk_data)
        frames["time"].append(float(dataset.getAttribute("timestep")))

    return frames


def read_time_series(filename):
    if filename.endswith(".vtk.series"):
        return read_vtk_time_series(filename)
    elif filename.endswith(".pvd"):
        return read_pvd_time_series(filename)
    else:
        raise IOError("Unknown file format of file %s" % filename)


def extract_field_data(vtk_data):
    origin_pointdata = np.array(vtk_data[0]["origin"])
    spacing = np.array(vtk_data[0]["spacing"])
    origin_celldata = origin_pointdata + 0.5 * spacing

    out_data = {}
    for point_data in vtk_data[0]["point_data"]:
        out_data[point_data] = {"origin": origin_pointdata, "spacing": spacing, "data": []}

    for cell_data in vtk_data[0]["cell_data"]:
        out_data[cell_data] = {"origin": origin_celldata, "spacing": spacing, "data": []}

    for data in vtk_data:
        if not np.allclose(data["origin"], origin_pointdata):
            raise IOError("Inconsistent origins within file series")
        if not np.allclose(data["spacing"], spacing):
            raise IOError("Inconsistent spacings within file series")

        for point_data, array in data["point_data"].items():
            out_data[point_data]["data"].append(array)

        for cell_data, array in data["cell_data"].items():
            out_data[cell_data]["data"].append(array)

    return out_data


def get_first_field(field_data):
    if len(list(field_data.keys())) > 1:
        warnings.warn("Results ambiguous since multiple fields inside vtk file")
    return field_data[list(field_data.keys())[0]]


def import_vtkdata(phia_vtk, phib_vtk, phic_vtk):
    phia = read_time_series(phia_vtk)
    phia_field = get_first_field(extract_field_data(phia["data"]))

    origin = phia_field["origin"]
    spacing = phia_field["spacing"]
    vtkdata = {"time": phia["time"], "phia": phia_field["data"]}

    vtks = {"phib": phib_vtk, "phic": phic_vtk}
    for phase, filename in vtks.items():
        time_series = read_time_series(filename)
        if not np.allclose(time_series["time"], vtkdata["time"]):
            raise IOError("Inconsistent times among phases")

        field = get_first_field(extract_field_data(time_series["data"]))

        if not np.allclose(field["origin"], origin):
            raise IOError("Inconsistent origins among phases")

        if not np.allclose(field["spacing"], spacing):
            raise IOError("Inconsistent spacings among phases")

        vtkdata[phase] = field["data"]

    vtkdata["grid"] = {
        "delta_x": spacing[0],
        "delta_y": spacing[1],
        "origin_x": origin[0],
        "origin_y": origin[1],
    }
    return vtkdata


def load_simdata(vtkdata, simparams_json):
    sim_params = load_sim_params_from_json(simparams_json)
    simdata = vtkdata.copy()
    simdata["time"] = [t * sim_params["time_unit"] for t in vtkdata["time"]]
    simdata["grid"] = {
        "delta_x": vtkdata["grid"]["delta_x"] * sim_params["length_unit"],
        "delta_y": vtkdata["grid"]["delta_y"] * sim_params["length_unit"],
        "origin_x": vtkdata["grid"]["origin_x"] * sim_params["length_unit"],
        "origin_y": vtkdata["grid"]["origin_y"] * sim_params["length_unit"],
    }
    simdata["input_params"] = sim_params

    return simdata



import json
from triplepy import ureg, Quantity
from pint import PintError
from triplepy.postprocessing import calc_p, calc_p_pf


def load_default_input_params(
    horizontal_resolution: int = 100, n_cells_int: float = 10.0
):
    simulation_params_default = {
        "gb_mobility": ureg("2e-8 m^4/(J*s)"),
        "gb_energy_horizontal": ureg("0.5 J/m^2"),
        "gb_energy_vertical": ureg("0.5 J/m^2"),
        "driving_force": ureg("1.0e5 Pa"),
        "width": ureg("50.0 um"),
        "horizontal_resolution": horizontal_resolution,
        "n_cells_interface": n_cells_int,
        "time_unit": ureg("1.0 s"),
        "length_unit": ureg("1.0 um").to("m")
    }
    return simulation_params_default


def update_driving_force(sim_params, p_value: float):
    out_params = sim_params.copy()
    p = p_value
    width = sim_params["width"]
    gb_energy_0 = sim_params["gb_energy_horizontal"]
    driving_force = (p * gb_energy_0) / width
    out_params["driving_force"] = driving_force.to("Pa")
    return out_params


def extract_numerical_params(params, unit_system):
    numerical_params = {}
    for parameter in params:
        quantity = params[parameter]
        if isinstance(quantity, Quantity):
            numerical_params[parameter] = quantity.to_preferred(unit_system).magnitude
        else:
            numerical_params[parameter] = quantity
    return numerical_params


def write_sim_params_to_json(sim_params, json_file_path: str, precision: int = 10):
    format_str = "%." + str(precision) + "g"

    out_params = sim_params.copy()
    out_params["p"] = calc_p(sim_params)
    out_params["p_pf"] = calc_p_pf(sim_params)

    # get rid of rounding errors from double precision
    for key in out_params:
        value = out_params[key]
        if isinstance(value, float):
            out_params[key] = float(format_str % out_params[key])

    # format pint quantities on the fly
    def json_formatter(obj):
        if isinstance(obj, Quantity):
            preferred_units = [ureg.J, ureg.m, ureg.s]
            obj = obj.to_preferred(preferred_units)
            value_str = format_str % obj.magnitude
            return value_str + f" {obj.units:~}"
        else:
            raise TypeError(
                f"Object of type {obj.__class__.__name__} is not JSON serializable"
            )

    with open(json_file_path, "w") as json_file:
        json.dump(out_params, json_file, default=json_formatter, indent=4)


def _is_float(obj: str) -> bool:
    try:
        float(obj)
        return True
    except ValueError:
        return False


def load_sim_params_from_json(json_file_path: str):
    with open(json_file_path, "r") as json_file:
        content = json.load(json_file)

    sim_params = {}
    for key, value in content.items():
        # try to convert a string value to a pint object
        if isinstance(value, str):
            # check if str starts with a number
            if _is_float(value.strip().split()[0]):
                # convert to pint object if possible
                try:
                    sim_params[key] = ureg(value)
                except PintError:
                    sim_params[key] = value
        else:
            sim_params[key] = value
    return sim_params



import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
import numpy as np


def plotField2D(field, title, dpi=100):
    [nx, ny] = field.shape
    plt.figure(figsize=(5, 5), dpi=dpi)
    plt.imshow(
        np.transpose(field), cmap="viridis", origin="lower", extent=[0, nx, 0, ny]
    )
    plt.xlabel("X")
    plt.ylabel("Y")
    plt.title(title)
    plt.show()


def plotInteractiveField2D(field, title):
    [nx, ny] = field[0].shape

    # Create the initial plot
    fig, ax = plt.subplots()
    im = ax.imshow(np.transpose(field[0]), cmap='viridis', origin='lower', extent=[0, nx, 0, ny])

    # Add a slider for changing timeframes
    frame_count = field.shape[0]-1
    ax_slider = plt.axes([0.2, 0.02, 0.6, 0.03])
    time_slider = Slider(ax_slider, 'Frame', 0, frame_count, valinit=0, valstep=1)

    # Function to update the plot based on the slider value
    def update(val):
        t = int(time_slider.val)
        im.set_array(field[t].T)
        ax.set_title(title+f' at time: {t}')
        fig.canvas.draw_idle()

    # Connect the slider to the update function
    time_slider.on_changed(update)

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    return time_slider  # Return the slider object


def plotTriplePoint(phia, phib, phic, dpi=200, grid=-1):
    plt.figure(figsize=(8, 8), dpi=dpi)
    if grid == -1:
        print("If you would like to specify the grid coordinates add grid=[X,Y] to function call.")
        [nx, ny] = phia.shape
        plt.imshow(np.transpose(phia+2*phib), cmap='viridis', origin='lower', extent=[0, nx, 0, ny])
        plt.contour(np.transpose(phia - phib), levels=[0.0], colors=['red'])
        plt.contour(np.transpose(phia - phic), levels=[0.0], colors=['blue'])
        plt.contour(np.transpose(phic - phib), levels=[0.0], colors=['white'])
    else:
        plt.imshow(np.transpose(phia+2*phib), cmap='viridis', origin='lower', extent=[grid[0][0, 0], grid[0][-1, -1], grid[1][0, 0], grid[1][-1, -1]])
        plt.contour(grid[0], grid[1], np.transpose(phia - phib), levels=[0.0], colors=['red'])
        plt.contour(grid[0], grid[1], np.transpose(phia - phic), levels=[0.0], colors=['blue'])
        plt.contour(grid[0], grid[1], np.transpose(phic - phib), levels=[0.0], colors=['white'])

    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Triple point visualization')
    plt.show()


def plotInteractiveTriplePoint(phia, phib, phic, dpi=200, m=1, f=1):
    global iso_ab
    [nx, ny] = phia[0].shape

    # Create the initial plot
    fig, ax = plt.subplots()
    im = ax.imshow(np.transpose(phia[0]+2*phib[0]), cmap='viridis', origin='lower', extent=[0, nx, 0, ny])
    ax.contour(np.transpose(phia[0] - phib[0]), levels=[0.0], colors=['red'])
    ax.contour(np.transpose(phia[0] - phic[0]), levels=[0.0], colors=['blue'])
    ax.contour(np.transpose(phic[0] - phib[0]), levels=[0.0], colors=['white'])

    # Add a slider for changing timeframes
    frame_count = phia.shape[0]-1
    ax_slider = plt.axes([0.2, 0.02, 0.6, 0.03])
    time_slider = Slider(ax_slider, 'Frame', 0, frame_count, valinit=0, valstep=1)

    # Function to update the plot based on the slider value
    def update(val):
        t = int(time_slider.val)
        im.set_array((phia[t]+2*phib[t]).T)
        for coll in ax.collections:
            coll.remove()

        ax.contour(np.transpose(phia[t] - phib[t]), levels=[0.0], colors=['red'])
        ax.contour(np.transpose(phia[t] - phic[t]), levels=[0.0], colors=['blue'])
        ax.contour(np.transpose(phic[t] - phib[t]), levels=[0.0], colors=['white'])
        ax.set_title(f'Triple point at frame: {t}')
        fig.canvas.draw_idle()

    # Connect the slider to the update function
    time_slider.on_changed(update)

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    return time_slider  # Return the slider object



import matplotlib.pyplot as plt
import numpy as np
import triplepy.sharp_interface_solution as si
from triplepy.numerical_evaluation import write_dict_to_csv

angle_deg = 120.0
m = si.calculate_slope_from_dihedral_angle(np.deg2rad(angle_deg))

driving_forces = [-100.0, -10.0, -1.0, 0.0, 1.0, 10.0, 100.0]
result_dict = {}
result_dict["x"] = np.linspace(0.0, 1.0, 1001)

fig, ax1 = plt.subplots()
for p in driving_forces:
    solver = si.GB_GeometrySolver(m, p)
    geometry = solver.calc_dimensionless_geometry(kind="double", relative_l2_tolerance=1e-8)
    key = "y_p%s" % (str(p))
    result_dict[key] = np.interp(result_dict["x"], geometry["x"], geometry["y"])
    ax1.plot(result_dict["x"], result_dict[key], "-", label="p = %.3f" % (p), lw=0.25)


ax1.set_aspect(1)
ax1.set_xlim(0.0, 1.0)
ax1.set_ylim(ax1.get_ylim()[0], 0.0)
ax1.set_xlabel("x")
ax1.set_ylabel("y")
plt.legend(loc="lower center", bbox_to_anchor=(0.5, 1.1), ncol=4)
plt.savefig("plot_geometries_theta_%g.pdf" % angle_deg, bbox_inches="tight")

write_dict_to_csv(result_dict, "analytical_geometries_theta_%g.csv" % angle_deg)



import pytest
import numpy as np
import os

import triplepy.vtk_io as vtk
import triplepy.postprocessing as pp

from triplepy.set_input_parameters import load_sim_params_from_json


@pytest.fixture
def root_folder():
    return os.path.join(os.path.dirname(__file__), "../data")


def test_import_vtk(root_folder):
    vtkdata = vtk.import_vtkdata(
        os.path.join(root_folder, "example_vtk_simulation/phia.vtk.series"),
        os.path.join(root_folder, "example_vtk_simulation/phib.vtk.series"),
        os.path.join(root_folder, "example_vtk_simulation/phic.vtk.series"),
    )
    simdata = vtk.load_simdata(
        vtkdata=vtkdata,
        simparams_json=os.path.join(
            root_folder, "example_vtk_simulation/simparams.json"
        ),
    )
    result = pp.postprocess_simulation(simdata)
    assert np.isclose(result["scalars"]["v_mean_phase"], -2.130934307)
    assert np.isclose(result["scalars"]["L2_mean"], 0.0434029331)

    lst_GB_sim = pp.calc_simulated_gb_profiles_in_frames(simdata=simdata, first_frame=39, last_frame=40)
    GB_sim = lst_GB_sim[0]
    assert GB_sim["frame_idx"] == 40
    assert np.isclose(GB_sim["L2"], 0.04371279725609258)


def test_import_pvd(root_folder):
    vtkdata = vtk.import_vtkdata(
        os.path.join(root_folder, "example_pvd_simulation/phia.pvd"),
        os.path.join(root_folder, "example_pvd_simulation/phib.pvd"),
        os.path.join(root_folder, "example_pvd_simulation/phic.pvd"),
    )
    simdata = vtk.load_simdata(
        vtkdata=vtkdata,
        simparams_json=os.path.join(
            root_folder, "example_pvd_simulation/simparams.json"
        ),
    )
    result = pp.postprocess_simulation(simdata)
    assert np.isclose(result["scalars"]["v_mean_phase"], -2.130934628)
    assert np.isclose(result["scalars"]["L2_mean"], 0.04387948137)

    lst_GB_sim = pp.calc_simulated_gb_profiles_in_frames(simdata=simdata, first_frame=39, last_frame=40)
    GB_sim = lst_GB_sim[0]
    assert GB_sim["frame_idx"] == 40
    assert np.isclose(GB_sim["L2"], 0.04341047094847948)


def import_vtkdata_micress(res_dir_path, basename):
    # Define dictionary for phase fields
    vtkdata = {"phia": [], "phib": [], "phic": [], "time": []}
    timetable_path = os.path.join(res_dir_path, f"{basename}_VTK_Time.txt")

    timetable = np.loadtxt(timetable_path)
    index_list = timetable[:, 0].astype(int)
    time_list = timetable[:, 1]

    first_file = os.path.join(res_dir_path, f"{basename}_{index_list[0]}.vtk")
    header = vtk.read_vtk_header(first_file)

    vtkdata["grid"] = {
        "delta_x": header["spacing"][0],
        "delta_y": header["spacing"][1],
        "origin_x": header["spacing"][0] * 0.5,
        "origin_y": header["spacing"][1] * 0.5,
    }

    for i, index in enumerate(index_list):
        vtk_file_path = os.path.join(res_dir_path, f"{basename}_{index}.vtk")
        field = vtk.read_vtk_field_data(vtk_file_path)
        # frac1 --> phia
        # frac2 --> phib
        # frac0 --> phic
        vtkdata["phia"].append(field["point_data"]["frac1"])
        vtkdata["phib"].append(field["point_data"]["frac2"])
        vtkdata["phic"].append(field["point_data"]["frac0"])
        vtkdata["time"].append(float(time_list[i]))

    vtkdata["phia"] = np.array(vtkdata["phia"])
    vtkdata["phib"] = np.array(vtkdata["phib"])
    vtkdata["phic"] = np.array(vtkdata["phic"])
    return vtkdata


def test_postprocess_micress_example(root_folder):
    res_dir_path = os.path.join(root_folder, "example_micress_simulation/Results")
    simparams_json = os.path.join(
        root_folder, "example_micress_simulation/simparams_0.json"
    )
    vtkdata = import_vtkdata_micress(
        res_dir_path=res_dir_path, basename="tripleBench_0"
    )
    simdata = vtk.load_simdata(vtkdata=vtkdata, simparams_json=simparams_json)
    results = pp.postprocess_simulation(simdata)
    # assert results["scalars"]["p"] == simdata["input_params"]["p"]
    assert np.isclose(results["scalars"]["v_analytic"], -11.79727890422015)
    assert np.isclose(results["scalars"]["v_mean_phase"], -11.822548437499998)
    assert np.isclose(results["scalars"]["L2_mean"], 0.008544205521747878)

    lst_GB_sim = pp.calc_simulated_gb_profiles_in_frames(simdata=simdata, first_frame=39, last_frame=40)
    GB_sim = lst_GB_sim[0]
    assert GB_sim["frame_idx"] == 40
    assert np.isclose(GB_sim["L2"], 0.008567135279800039)


def test_calc_p(root_folder):
    sim_params = load_sim_params_from_json(
        os.path.join(root_folder, "example_micress_simulation/simparams_0.json")
    )
    p = pp.calc_p(sim_params)
    assert p == sim_params["p"]


def test_calc_p_pf(root_folder):
    sim_params = load_sim_params_from_json(
        os.path.join(root_folder, "example_micress_simulation/simparams_0.json")
    )
    p_pf = pp.calc_p_pf(sim_params)
    assert p_pf == sim_params["p_pf"]



import triplepy.set_input_parameters as set_input


def test_updata_p():
    default_params = set_input.load_default_input_params()
    new_params = set_input.update_driving_force(default_params, 80)
    new_driving_force = new_params["driving_force"]
    assert new_driving_force.magnitude == 8.0e5



import numpy as np
from scipy import optimize
import warnings


def calculate_slope(gb_energy_moving, gb_energy_stationary):
    return 1.0 / np.sqrt(4.0 * (gb_energy_moving / gb_energy_stationary)**2 - 1)


def calculate_dihedral_angle(gb_energy_moving, gb_energy_stationary):
    energy_ratio = gb_energy_stationary / gb_energy_moving
    return 2.0 * np.arccos(0.5*energy_ratio)


def calculate_gb_energy_ratio_from_dihedral_angle(theta):
    return 2.0 * np.cos(0.5 * theta)


def calculate_dihedral_angle_from_slope(m):
    return np.pi - 2.0*np.arctan(m)


def calculate_slope_from_dihedral_angle(theta):
    return np.tan(0.5*(np.pi - theta))


def isclose(a, b):
    return np.isclose(a, b)


def calc_l2_norm_rel_error(x_coarse, f_coarse, x_fine, f_fine):
    f_coarse_on_fine = np.interp(x_fine, x_coarse, f_coarse)
    diff2 = (f_coarse_on_fine - f_fine)*(f_coarse_on_fine - f_fine)
    return np.sqrt(np.trapz(diff2, x=x_fine) / np.trapz(np.array(f_fine)*np.array(f_fine), x=x_fine))


def integrate_trapezoidal(y0, x, slopes):
    ys = [y0]
    for i in range(len(x)-1):
        # trapezodial rule
        ys.append(ys[-1] + (slopes[i] + slopes[i+1]) * 0.5 * (x[i+1] - x[i]))
    return x, np.array(ys)


def integrate(y0, x, slopes):
    return integrate_trapezoidal(y0, x, slopes)


class GB_VelocityCalculator:
    def __init__(self, triple_junction_slope):
        self.m = triple_junction_slope
        self.sqrt_1_m2 = np.sqrt(1.0 + self.m * self.m)
        self.atan_m = np.arctan(self.m)

        # p, where v = 0
        self.p_stat = 2.0*self.m / self.sqrt_1_m2

        # p, where v = -p
        self.p_crit_neg1 = -2.0 * ((self.sqrt_1_m2 - 1.0)/self.m - self.atan_m)

        self.v_crit_neg1_m2 = np.log(1.0 + self.m * self.m) / self.m - 2.0 * self.atan_m
        # p, where v = -sqrt(1+m^2) p
        self.p_crit_neg1_m2 = -self.v_crit_neg1_m2 / self.sqrt_1_m2

        # we should have the following order p_crit_neg1_m2 < p_crit_neg1 < p_stat
        # p_crit_neg1 should always be lower than p_stat
        if self.p_crit_neg1 > self.p_stat:
            raise ValueError("The dimensional driving force for stationary GB (p_stat=%f) is larger than the one with q = -1 (p_neg1=%f)" % (self.p_stat, self.p_crit_neg1))

        # p_crit_neg1_m2 should always be lower than p_crit_neg1
        if self.p_crit_neg1_m2 > self.p_crit_neg1:
            raise ValueError("The dimensional driving force for q=-1 (p_neg1=%f) is larger than the one with q=-sqrt(1+m^2) (p_neg1_m2=%f)" % (self.p_crit_neg1, self.p_crit_neg1_m2))

        self.__run_tests()

    # "The most effective debugging techniques seem to be those which are designed and built into the program itself."
    # Donald E. Knuth
    # From Dr. Dobb's Journal, an article on debugging Windows programs, November 1993.
    def __run_tests(self):
        self.__test_velocity_neg1()
        self.__test_velocity_neg1_m2()

    def __test_velocity_neg1(self):
        if not isclose(self.calculate_velocity(self.p_crit_neg1), -self.p_crit_neg1):
            raise ValueError("The velocity for q=-1 (p_neg1=%f) is not identical to -p_neg1." % (self.p_crit_neg1))

    def __test_velocity_neg1_m2(self):
        if not isclose(self.calculate_velocity(self.p_crit_neg1_m2), self.v_crit_neg1_m2):
            raise ValueError("The velocity for q=-sqrt(1+m^2) (p_neg1_m2=%f) is not identical to -sqrt(1+m^2)*p_neg1_m2." % (self.p_crit_neg1_m2))

    def __calculate_velocity_lower_bound(self, p):
        # return lower bounds with decreasing order of p
        if p > self.p_stat:  # velocity is positive for p > p_stat
            return -1e-12
        if p > self.p_crit_neg1:  # velocity is larger -p_crit_neg1 for p_crit_neg1 < p < p_stat
            return -self.p_crit_neg1
        if p > self.p_crit_neg1_m2:  # velocity is larger v_crit_neg1_m2 for p_crit_neg1_m2 < p < p_crit_neg1
            return self.v_crit_neg1_m2
        if p > 0.0:  # velocity is larger -2 atan(m) for 0.0 < p < p_crit_neg1_m2
            return -2.0*self.atan_m
        # velocity is larger than this lower bound for negative p
        return np.minimum(-6.0*self.atan_m, -np.sqrt((1.0 + self.m*self.m)*p*p + 6.0*self.m*np.abs(p)))

    def __calculate_velocity_upper_bound(self, p):
        # return upper bounds with decreasing order of p
        if p > self.p_stat:  # velocity is smaller p for p > p_stat ( we add some correction since otherwise the root func is undefined )
            return p * (1.0 - 1.0e-12)
        if p > self.p_crit_neg1:  # velocity is below 0.0 for p_crit_neg1 < p < p_stat
            return 1e-12
        if p > self.p_crit_neg1_m2:  # velocity is below -p_crit_neg1 for p_crit_neg1_m2 < p < p_crit_neg1
            return -self.p_crit_neg1
        if p > 0.0:  # velocity is below v_crit_neg1_m2 for 0.0 < p < p_crit_neg1_m2
            return self.v_crit_neg1_m2
        # velocity is below -2.0*self.atan_m and below self.sqrt_1_m2*p for negative p
        return np.minimum(-2.0*self.atan_m, self.sqrt_1_m2*p*(1.0 - 1e-12))  # ( we add some correction since otherwise the root func is undefined )

    def __rootfunc(self, v, p):
        sqrt_vp = np.emath.sqrt(v*v - p*p)
        factor_1 = p / (v*sqrt_vp)
        # keep in mind that arctan(x/i)/i = -arctanh(x)
        return -np.real(factor_1 * (np.emath.arctanh(self.m*v / (self.sqrt_1_m2*sqrt_vp)) + np.emath.arctanh(self.m*p / sqrt_vp))) - self.atan_m/v - 0.5

    def __solve_single_velocity(self, p):
        if isclose(p, 0.0):
            return -2.0*self.atan_m

        if isclose(p, self.p_stat):
            return 0.0

        if isclose(p, self.p_crit_neg1):
            return -self.p_crit_neg1

        if isclose(p, self.p_crit_neg1_m2):
            return self.v_crit_neg1_m2

        lower_bound = self.__calculate_velocity_lower_bound(p)
        upper_bound = self.__calculate_velocity_upper_bound(p)
        lower_rootfunc = self.__rootfunc(lower_bound, p)
        upper_rootfunc = self.__rootfunc(upper_bound, p)

        # In case we have large driving forces, the true upper/lower bound cannot be used because of an undefined root function.
        # Since we correct with a relative factor of 1e-12 it may happen that the signs of the root function are not different.
        # This can only happen if the real velocity is above the estimated upper bound or below the estimated lower bound which means it has to be very near (< 1e-12 relative error) the linear regime.
        if np.sign(lower_rootfunc) == np.sign(upper_rootfunc):
            # Make the correction for large driving forces, leave unresolved otherwise which will raise an error in the brenth routine automatically.
            if p < 0.0:
                return self.sqrt_1_m2 * p
            elif p > self.p_stat:
                return p

        root = optimize.brenth(self.__rootfunc, lower_bound, upper_bound, args=(p))
        return root

    def calculate_velocity(self, p):
        if (isinstance(p, float) or isinstance(p, int)):
            return self.__solve_single_velocity(p)

        res = []
        for p_i in p:
            res.append(self.__solve_single_velocity(p_i))
        return np.array(res)


class GB_GeometrySolver:
    def __init__(self, triple_junction_slope, driving_force):
        calculator = GB_VelocityCalculator(triple_junction_slope)
        self.m = triple_junction_slope
        self.v = calculator.calculate_velocity(driving_force)
        self.p = driving_force

    def __xfunction(self, slope):
        v = self.v
        p = self.p
        # simple solution without driving force, exclude to allow dividing by p
        if isclose(p, 0.0):
            return np.arctan(slope) / v

        # exclude limiting special cases where denominator vanishes
        if isclose(v, 0.0):
            return -slope / (p * np.sqrt(1.0 + slope * slope))

        if isclose(v/p, -1.0):
            return np.where(isclose(slope, 0.0), 0.0, ((1.0 - np.sqrt(1.0 + slope * slope)) / slope + np.arctan(slope)) / v)

        sqrt_vp = np.emath.sqrt(v*v - p*p)  # can become imaginary
        result = np.full(slope.shape, np.nan)
        result[slope == self.m] = -0.5
        result[slope == -self.m] = 0.5
        mask = np.abs(slope) != self.m
        sqrt_slope = np.sqrt(1.0 + slope[mask]*slope[mask])
        # keep in mind that arctan(x/i)/i = -arctanh(x)
        result[mask] = np.real(p/(v*sqrt_vp) * (np.emath.arctanh((p*slope[mask])/sqrt_vp) + np.emath.arctanh((v*slope[mask])/(sqrt_vp * sqrt_slope)))) + np.arctan(slope[mask]) / v
        return result

    def __generate_slope_samples(self, n_samples):
        # generate slopes from 0 to -m
        slopes_intermediate = np.linspace(0.0, -0.9*self.m, n_samples)
        slopes_near_tj = np.concatenate((-self.m*(1.0-np.geomspace(0.1, 1e-9, n_samples))[1:], [-self.m]))
        return np.concatenate((slopes_intermediate, slopes_near_tj))

    def calc_dimensionless_geometry(self, relative_l2_tolerance=1.0e-8, max_refinements=100, kind="double"):
        # we start with only four samples in the beginning
        n_samples = 2
        slopes = self.__generate_slope_samples(n_samples)

        # calculate corresponding dimensionless x-coordinates
        xtilde = self.__xfunction(slopes)

        error = 0.0
        it = 0
        while ((it == 0 or error > relative_l2_tolerance) and it < max_refinements):
            it += 1

            x, y = integrate(0.0, xtilde, slopes)
            n_samples *= 2

            slopes_fine = self.__generate_slope_samples(n_samples)

            xtilde_fine = self.__xfunction(slopes_fine)

            x_fine, y_fine = integrate(0.0, xtilde_fine, slopes_fine)

            error = calc_l2_norm_rel_error(x, y, x_fine, y_fine)

            xtilde = xtilde_fine
            slopes = slopes_fine

        if error <= relative_l2_tolerance:
            print("Solution of GB geometry converged with L2 error = %.1e after %d refinements" % (error, it))
        else:
            warnings.warn("Solution did not converge! L2 error = %.1e" % error)

        # produce appropriate output for "kind"
        if kind == "half":
            return {"x": x_fine, "y": y_fine, "derivative": slopes_fine}
        if kind == "full":
            return {"x": np.concatenate((-np.flip(x_fine)[:-1], x_fine)),
                    "y": np.concatenate((np.flip(y_fine)[:-1], y_fine)),
                    "derivative": np.concatenate((-np.flip(slopes_fine)[:-1], slopes_fine))
                    }
        if kind == "double":
            # remove discontinuity of slope at triple junction
            slopes_fine[-1] = 0.0
            return {"x": np.concatenate((x_fine, 1.0 - np.flip(x_fine[:-1]))),
                    "y": np.concatenate((y_fine, np.flip(y_fine[:-1]))),
                    "derivative": np.concatenate((slopes_fine, -np.flip(slopes_fine[:-1])))
                    }
        else:
            raise ValueError("Unknown kind \"%s\"" % kind)



import numpy as np
import triplepy.sharp_interface_solution as si


def test_angle_slope():
    theta = np.deg2rad([45.0, 90.0, 110.0, 145.0])
    m = si.calculate_slope_from_dihedral_angle(theta)
    assert np.allclose(si.calculate_dihedral_angle_from_slope(m), theta)


def test_ratio_angle():
    gamma_ratio = np.array([0.5, 1.0, 1.5, 1.8])
    angle = si.calculate_dihedral_angle(1.0, gamma_ratio)
    assert np.allclose(si.calculate_gb_energy_ratio_from_dihedral_angle(angle), gamma_ratio)


def test_ratio_slope():
    gamma_ratio = np.array([0.5, 1.0, 1.5, 1.8])
    m = si.calculate_slope(1.0, gamma_ratio)
    angle = si.calculate_dihedral_angle_from_slope(m)
    assert np.allclose(si.calculate_gb_energy_ratio_from_dihedral_angle(angle), gamma_ratio)



import numpy as np

import triplepy.numerical_evaluation as numerics
import triplepy.sharp_interface_solution as analytic


def calc_p(sim_params):
    p = (
        sim_params["driving_force"]
        * sim_params["width"]
        / sim_params["gb_energy_horizontal"]
    )
    return float(p)


def calc_m(sim_params):
    m = analytic.calculate_slope(
        sim_params["gb_energy_horizontal"], sim_params["gb_energy_vertical"]
    )
    return float(m)


def calc_p_pf(sim_params):
    delta_x = sim_params["width"] / sim_params["horizontal_resolution"]
    eta = delta_x * sim_params["n_cells_interface"]
    p_pf = sim_params["driving_force"] * eta / sim_params["gb_energy_horizontal"]
    return round(float(p_pf), 10)


def calc_transient_time(sim_params):
    transient_time = (sim_params["width"]) ** 2 / (
        sim_params["gb_mobility"] * sim_params["gb_energy_horizontal"]
    )
    return transient_time


def calc_velocity_dimensional(v_dimless, sim_params):
    mobility = sim_params["gb_mobility"]
    gb_energy = sim_params["gb_energy_horizontal"]
    width = sim_params["width"]
    velocity = (v_dimless * mobility * gb_energy / width).to_base_units()
    return velocity


def calc_analytic_velocity_nondimensional(m, p):
    # Compute (non-dimensional) analytical velocity
    gb = analytic.GB_VelocityCalculator(m)
    v_analytic = gb.calculate_velocity(p)
    return v_analytic


def calc_dimless_velocity_in_frames(
    simdata, first_frame: int = 1, last_frame: int = None
):
    v_left = []
    v_right = []
    v_triple = []
    v_phase = []
    if last_frame is None:
        last_frame = simdata["phia"].shape[0] - 1

    # Note that in the micress vtk data, the first frame are at time 0
    # thus we start with frame 1
    lst_frame_idx = []
    for frame in range(first_frame + 1, last_frame + 1):
        v = numerics.calculate_velocity_nondimensional(simdata, frame - 1, frame)
        v_left.append(v["left_boundary"])
        v_right.append(v["right_boundary"])
        v_triple.append(v["triple_point"])
        v_phase.append(v["phase_fraction"])
        lst_frame_idx.append(frame)
    v_dimless = {}
    v_dimless["frame_indices"] = np.array(lst_frame_idx)
    v_dimless["v_left"] = np.array(v_left)
    v_dimless["v_right"] = np.array(v_right)
    v_dimless["v_triple"] = np.array(v_triple)
    v_dimless["v_phase"] = np.array(v_phase)
    return v_dimless


def calc_analytic_grain_boundary(m, p):
    solver = analytic.GB_GeometrySolver(m, p)
    GB_analytic = solver.calc_dimensionless_geometry(
        kind="double", relative_l2_tolerance=1e-7
    )

    GB_x_analytic = GB_analytic["x"] - np.min(GB_analytic["x"])
    GB_y_analytic = GB_analytic["y"] - np.max(GB_analytic["y"])
    GB_height_analytic = np.max(GB_y_analytic) - np.min(GB_y_analytic)
    return GB_x_analytic, GB_y_analytic, GB_height_analytic


def calc_simulated_gb_profiles_in_frames(
    simdata, first_frame: int = 1, last_frame: int = None
):
    if last_frame is None:
        last_frame = simdata["phia"].shape[0] - 1

    # Compute analytical GB profile
    m = calc_m(simdata["input_params"])
    p = calc_p(simdata["input_params"])
    GB_x_analytic, GB_y_analytic, GB_height_analytic = calc_analytic_grain_boundary(
        m, p
    )

    # Evaluate simulated profile in frames
    GB_sim_in_frames = []
    for i, frame in enumerate(range(first_frame + 1, last_frame + 1)):
        if i == 0:
            _, GB_y_sim_old = numerics.extract_grainboundary_profile(simdata, frame - 1)
            height_old = np.max(GB_y_sim_old) - np.min(GB_y_sim_old)
        else:
            height_old = GB_sim_in_frames[i - 1]["height"]
        GB_x_sim, GB_y_sim = numerics.extract_grainboundary_profile(simdata, frame)
        GB_sim = {}
        GB_sim["frame_idx"] = frame
        GB_sim["x"] = GB_x_sim
        GB_sim["y_sim"] = GB_y_sim
        GB_sim["height"] = np.max(GB_y_sim) - np.min(GB_y_sim)
        # Quantify convergence through change in GB profile height
        GB_sim["rel_height_change"] = (
            GB_sim["height"] - height_old
        ) / GB_height_analytic
        # Quantify relative error in GB profile height
        GB_sim["rel_height_error"] = (
            GB_sim["height"] - GB_height_analytic
        ) / GB_height_analytic
        # Quantify deviation of GB profile in terms of L2-Norm
        if GB_x_sim.size > 1:
            GB_y_analytic_interp = np.interp(GB_x_sim, GB_x_analytic, GB_y_analytic)
            diff2 = (GB_y_sim - GB_y_analytic_interp) ** 2
            GB_sim["L2"] = np.sqrt(
                np.trapz(diff2, x=GB_x_sim)
                / np.trapz(GB_y_analytic_interp**2, x=GB_x_sim)
            )
        else:
            GB_y_analytic_interp = np.nan
            GB_sim["L2"] = np.nan
        GB_sim["y_analytic_interp"] = GB_y_analytic_interp
        GB_sim_in_frames.append(GB_sim)
    return GB_sim_in_frames


def postprocess_simulation(
    simdata,
    first_frame: int = None,
    last_frame: int = None,
    display_message: bool = True,
):
    """
    Postprocesses simulation data to evaluate and summarize simulation results.

    Args:
        simdata (dict): Simulation data containing grid info, input parameters, 2D field values, and time series.
            The structure of simdata is as follows:
            - "grid": dict with grid information like dx, dy, and origin of first grid point/cell center.
            - "input_params": dict of physical simulation input parameters; can be read from a JSON file.
            - "phia", "phib", "phic": time-series of 2D field values saved as numpy arrays
              with dimensions (number of frames, points in x, points in y).
            - "time": list of timeframes.
        first_frame (int, optional): Index of the first frame to consider for analysis. Defaults to None.
        last_frame (int, optional): Index of the last frame to consider for analysis. Defaults to None.
        display_message (bool, optional): Flag to control the display of messages. Defaults to True.

    Raises:
        ValueError: If the first_frame is greater than or equal to the last_frame.

    Returns:
        dict: A dictionary containing geometry data and scalar results.
    """
    # determine evaluation frame range
    # From dimensionality analysis we compute the following transient time
    transient_time = calc_transient_time(simdata["input_params"])

    # The simulation should at least be run the transient time
    # to ensure convergence of the system.
    dimless_times = np.array([float(t / transient_time) for t in simdata["time"]])
    if (first_frame is None) or (last_frame is None):
        frame_start = np.where(dimless_times >= 0.5)[0][0]
        frame_end = np.where(dimless_times == 1.0)[0][0]
    elif first_frame >= last_frame:
        raise ValueError(
            "The first frame number should be smaller than the last frame number."
        )
    else:
        frame_start = first_frame
        frame_end = last_frame

    # Compute analytical velocity
    # Compute slope from Young's law
    m = calc_m(simdata["input_params"])
    # Compute non-dimensional driving force
    p = calc_p(simdata["input_params"])
    v_analytic = calc_analytic_velocity_nondimensional(m, p)
    if display_message:
        print(f"TJ slope = {m:g}")
        print(f"Dimensionless driving force = {p:g}")
        print(f"Analytic velocity = {v_analytic:g}")
        print("\n")

    # Evaluate velocities in frames
    v_dimless = calc_dimless_velocity_in_frames(
        simdata=simdata, first_frame=frame_start, last_frame=frame_end
    )
    v_left = v_dimless["v_left"]
    v_phase = v_dimless["v_phase"]

    # Evaluation of gb profiles
    GB_sim_in_frames = calc_simulated_gb_profiles_in_frames(
        simdata=simdata, first_frame=frame_start, last_frame=frame_end
    )
    # Quantify deviations
    lst_L2 = [GB_sim["L2"] for GB_sim in GB_sim_in_frames]
    lst_L2 = np.array(lst_L2)
    lst_rel_height_change = np.array(
        [GB_sim["rel_height_change"] for GB_sim in GB_sim_in_frames]
    )
    lst_rel_height_error = np.array(
        [GB_sim["rel_height_error"] for GB_sim in GB_sim_in_frames]
    )

    # final geometry
    GB_sim_final = GB_sim_in_frames[-1]
    geometry_dict = {}
    geometry_dict["x"] = GB_sim_final["x"]
    geometry_dict["y_sim"] = GB_sim_final["y_sim"]
    geometry_dict["y_analytic"] = GB_sim_final["y_analytic_interp"]

    results = {
        "geometries": geometry_dict,
        "scalars": {
            "p": p,
            "v_analytic": v_analytic,
            "v_mean_left": np.mean(v_left),
            "v_std_left": np.std(v_left),
            "v_mean_phase": np.mean(v_phase),
            "v_std_phase": np.std(v_phase),
            "height_change_mean": np.mean(lst_rel_height_change),
            "height_change_std": np.std(lst_rel_height_change),
            "height_error_mean": np.mean(lst_rel_height_error),
            "height_error_std": np.std(lst_rel_height_error),
            "L2_mean": np.mean(lst_L2),
            "L2_std": np.std(lst_L2),
        },
    }
    return results


def make_results_table(simulations, keys):
    table = {key: [] for key in keys}
    for sim in simulations:
        for key in keys:
            table[key].append(sim[key])
    return table



import triplepy.sharp_interface_solution as si

calc = si.GB_GeometrySolver(3.0, 5.0)
# This is our reference solution
geometry = calc.calc_dimensionless_geometry(relative_l2_tolerance=1.0e-8)
print("Reference resolution = %i" % len(geometry["x"]))

for tolerance in [1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6]:
    print("-------------------------")
    coarse_geometry = calc.calc_dimensionless_geometry(relative_l2_tolerance=tolerance)
    l2_error = si.calc_l2_norm_rel_error(coarse_geometry["x"], coarse_geometry["y"], geometry["x"], geometry["y"])
    print("Resolution = %i, Tolerance = %.2e, L2_error = %.2e, error_ratio = %.2g" % (len(coarse_geometry["x"]), tolerance, l2_error, l2_error / tolerance))
    print("Accuracy goal passed? %s" % "YES" if l2_error < tolerance else "NO")
